package info.heinzelnisse.he;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.net.URL;
import java.text.CollationKey;
import java.text.Collator;
import java.text.ParseException;
import java.text.RuleBasedCollator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Title:        Heinzelnisse.HeinzelTable
 * Description:  Desktop Edition of Heinzelnisse
 * Copyright:    Copyright (c) 2002
 * Company:
 * @author Heiko Klein
 * @version 1.6
 */
public class HeinzelTable implements Runnable, Serializable {
    public final static String CHARSET = "ISO8859-1";
    public final static String SEPARATOR = "\t";
    public final static String OPTIONAL_SEPARATOR = ";";
    public static final String HEINZELLISTE = "/heinzelliste.txt";
    public static final String FINISHED_READING = "finishedReading";
    public static final int MAX_DISPLAYABLE_ITEMS = 1000;
    public static final int MAX_HISTORY_ITEMS = 1000;
    
    protected final PropertyChangeSupport pcs = new PropertyChangeSupport(this);

    private static Logger logger = Logger.getLogger(HeinzelTable.class.getName());
    private boolean initialized = false;
    private static Collator deNoCollator = null;
    private Thread readThread = null;
    // cached sorted list of current items
    private volatile boolean currentIsNorwegian = true;
    private volatile boolean currentIsGerman = true;
    private volatile List<String> currentSortedItems = null;
    // contains norwegian words
    private Map<String, List<String>> norwegianTranslationItems = null;
    // contains german words
    private Map<String, List<String>> germanTranslationItems = null;
    private URL dataTable;

    public HeinzelTable() {
        this(info.heinzelnisse.he.HeinzelTable.class.getResource(HEINZELLISTE));
    }

    public HeinzelTable(URL url) {
        // german Collator covers everything, no need to join rules of german + norwegian
        dataTable = url;
    }

    private void serialize(String fileName, Map<String, List<String>> stringMap) throws IOException {
        OutputStreamWriter osw = new OutputStreamWriter(new BufferedOutputStream(new FileOutputStream(fileName)));
        for (String s : stringMap.keySet()) {
            List<String> list = stringMap.get(s);
            StringBuffer vals = new StringBuffer();
            for (Iterator<String> it = list.iterator(); it.hasNext();) {
                vals.append(it.next());
                if (it.hasNext()) {
                    vals.append("\u0000");
                }
            }
            osw.write(s);
            osw.write("\n");
            osw.write(vals.toString());
            osw.write("\n");
        }
        osw.close();
    }


    public void addPropertyChangeListener(PropertyChangeListener listener) {
        this.pcs.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener) {
        this.pcs.removePropertyChangeListener(listener);
    }

    public synchronized void run() {
        readTable(dataTable);
        //readFromSerialized();
    }

    private synchronized void readTable(URL url) {
        BufferedReader d = null;
        try {
            logger.info("Start reading: " + url.toString());
        } catch (Exception e) {
            logger.warning("error in input:" + e.toString());
        }
        try {
            d = new BufferedReader(new InputStreamReader(new BitObfuscatorInputStream(new BufferedInputStream(url.openStream())), CHARSET));
        } catch (Exception e) {
            e.printStackTrace();
        }
        if (d != null) {
            readTable(d);
        }
    }

    private synchronized void readTable(BufferedReader d) {
        // empyt all lists
        long start = System.currentTimeMillis();
        norwegianTranslationItems = new HashMap<String, List<String>>(40000);
        germanTranslationItems = new HashMap<String, List<String>>(40000);

        try {
            // matches word1 and word2, everything stays
            Pattern p = Pattern.compile("^([^"+SEPARATOR+"]+)"+SEPARATOR+"[^"+SEPARATOR+"]*"+SEPARATOR+"[^"+SEPARATOR+"]*"+SEPARATOR+"[^"+SEPARATOR+"]*"+SEPARATOR+"([^"+SEPARATOR+"]+)"+SEPARATOR);
            Matcher m = p.matcher(""); // init
            String line =d.readLine(); // skip header line
            while ((line = d.readLine()) != null) {
                m.reset(line);
                if (m.find()) {
                    addNorwegianSearchItem(m.group(1), line);
                    addGermanSearchItem(m.group(2), line);
                }
            }
            logger.info("SearchItems: " + (getNorwegianTranslationItems().size() + getGermanTranslationItems().size()));
            logger.info("Finished reading after: " + (System.currentTimeMillis() - start) + " ms");
            start = System.currentTimeMillis();
            getSortedItems();
            logger.info("Finished sorting after: " + (System.currentTimeMillis() - start) + " ms");

            boolean oldInit = isInitialized();
            setInitialized(true);
            this.pcs.firePropertyChange(FINISHED_READING, oldInit, isInitialized());
            //serialize();
        } catch (IOException e) {
            System.err.println(e.toString());
        }
    }

    public synchronized List<String> getSortedItems() {
        return getSortedItems(currentIsNorwegian, currentIsGerman);
    }

    public synchronized List<String> getSortedItems(boolean norwegian, boolean german) {
        if ((currentSortedItems != null) && (norwegian == this.currentIsNorwegian) && (german == this.currentIsGerman)) {
            return currentSortedItems;
        }

        // now calculate
        logger.info("new initialized sortItems: " + (norwegian ? "norwegian" : "") + (german ? "german" : ""));
        currentSortedItems = null;
        currentIsNorwegian = norwegian;
        currentIsGerman = german;
        int size = 0;
        if (norwegian) {
            size += getNorwegianTranslationItems().size();
        }
        if (german) {
            size += getGermanTranslationItems().size();
        }
        CollationKey[] collKeys = new CollationKey[size];
        int i = 0;
        if (norwegian) {
            for (String s : getNorwegianTranslationItems().keySet()) {
                collKeys[i++] = getDeNoCollator().getCollationKey(s);
            }
        }
        if (german) {
            for (String s : getGermanTranslationItems().keySet()) {
                collKeys[i++] = getDeNoCollator().getCollationKey(s);
            }
        }
        Arrays.sort(collKeys);
        currentSortedItems = new ArrayList<String>(size);
        for (CollationKey ck : collKeys) {
            currentSortedItems.add(ck.getSourceString());
        }
        return currentSortedItems;
    }

    private void addNorwegianSearchItem(String item, String translation) {
        addSearchItem(item, translation, getNorwegianTranslationItems());
    }

    private void addGermanSearchItem(String item, String translation) {
        addSearchItem(item, translation, getGermanTranslationItems());
    }

    private void addSearchItem(String item, String translation, Map<String, List<String>> langTranslationItems) {
        item = item.trim();
        List<String> rows = langTranslationItems.get(item);
        if (rows == null) {
            // new item
            rows = new ArrayList<String>(2);
            rows.add(translation);
            langTranslationItems.put(item, rows);
        } else if (!translation.equals(rows.get(rows.size() - 1))) {
            rows.add(translation);
            langTranslationItems.put(item, rows);
        }
    }

    public synchronized List<String> searchItems(String str) {
        return HeinzelSearch.items(str + HeinzelSearch.ANY, getSortedItems());
    }

    public synchronized List<String> searchTranslations(String str) {
        List<String> translations = new ArrayList<String>();
        Set<String> rows = new HashSet<String>();
        List<String> items = searchItems(str);
        logger.log(Level.INFO, "found {0} items for {1}", new Object[]{items.size(), str});
        for (String item : items) {
            logger.log(Level.INFO, "found item: {0}", item);
            if (currentIsNorwegian) {
                List<String> transItem = getNorwegianTranslationItems().get(item);
                if (transItem != null) {
                    rows.addAll(transItem);
                }
            }
            if (currentIsGerman) {
                List<String> transItem = getGermanTranslationItems().get(item);
                if (transItem != null) {
                    rows.addAll(transItem);
                }
            }
        }
        if (rows != null) {
            for (String s : rows) {
                translations.add(s);
            }
        }
        return translations;
    }

    public synchronized Collection<String> searchFulltextTranslations(String str) {
        return searchFulltextTranslations(str, false, true, currentIsNorwegian, false, currentIsNorwegian, currentIsNorwegian, currentIsGerman, false, currentIsGerman, currentIsGerman, false);
    }

    public synchronized Collection<String> searchFulltextTranslations(String str, boolean caseSensitive, boolean subString, boolean wordNO, boolean articleNO, boolean optionsNO, boolean otherNO, boolean wordDE, boolean articleDE, boolean optionsDE, boolean otherDE, boolean category) {
        List<String> results = new ArrayList<String>();
        String wordString;
        if (subString) {
            wordString = ".*" + str + ".*";
        } else {
            wordString = str;
        }
        Pattern wordPattern;
        Pattern pattern;
        if (caseSensitive) {
            pattern = Pattern.compile(str);
            wordPattern = Pattern.compile(wordString);
        } else {
            pattern = Pattern.compile(str, Pattern.CASE_INSENSITIVE);
            wordPattern = Pattern.compile(wordString, Pattern.CASE_INSENSITIVE);
        }
        Matcher matcher = pattern.matcher(""); // init
        Matcher wordMatcher = wordPattern.matcher("");
        Set<String> translations = new HashSet<String>();
        for (List<String> ls : getNorwegianTranslationItems().values()) {
            for (String s : ls) {
                matcher.reset(s);
                if (matcher.find()) {
                    // whole row contains word
                    logger.fine(s);
                    Translation trans = Translation.getTranslationOf(s, Translation.NO);
                    if (wordNO) {
                        wordMatcher.reset(trans.getWord());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (wordDE) {
                        wordMatcher.reset(trans.getTransWord());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (articleNO) {
                        wordMatcher.reset(trans.getArticle());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (articleDE) {
                        wordMatcher.reset(trans.getTransArticle());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (optionsNO) {
                        wordMatcher.reset(trans.getOptions());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (optionsDE) {
                        wordMatcher.reset(trans.getTransOptions());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (otherNO) {
                        wordMatcher.reset(trans.getOther());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (otherDE) {
                        wordMatcher.reset(trans.getTransOther());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                    if (category) {
                        wordMatcher.reset(trans.getCategory());
                        if (wordMatcher.matches()) {
                            translations.add(s);
                            continue;
                        }
                    }
                }
            }
        }
        return translations;
    }

    public synchronized boolean isInitialized() {
        return initialized;
    }

    public synchronized void setInitialized(boolean initialized) {
        this.initialized = initialized;
    }

    public static Collator getDeNoCollator() {
        if (deNoCollator == null) {
            try {
                deNoCollator = new RuleBasedCollator("<'\u0000'<' '< " +
                                                   "a;ä,A;Ä<b,B<c,C<d,D<e,E<f,F<g,G<h,H<i,I<j,J<k,K<l,L<m,M<n,N<o;ö;ó;O;Ö<p,P<q,Q<r,R<s, S & SS,ß<t,T<u;ü,U;Ü<v,V<w,W<x,X<y,Y<z,Z <"+
                                                   "æ=œ,AE=Æ < ø,Ø < å=Å=AA < '(' < '.'< '#' < '\"' < 0<1<2<3<4<5<6<7<8<9<«<» ");
//                deNoCollator = new RuleBasedCollator("'\u0000'<'-';­;‐;‑;‒;–;—;―;−<' '< '_' <¯<','<';'<':'<'!'<¡<'?'<'/'<´<'`'<'^'<'~'<·<¸<'('<')'<'['<']'<'{'<'}'<§<¶<©<®<'@'<¢<₡<₢<'$'<₫<€<₣<₤<£<¥<'*'<'\\'<'&'<'#'<'%'<'+'<±<÷<×<" +
//                                                   "'<'<'='<'>'<'|'<¦<°<µ<¼<½<¾<"+
//                                                   "a;ä,A;Ä<b,B<c,C<d,D<ð,Ð<e,E<f,F<g,G<h,H<i,I<j,J<k,K<l,L<m,M<n,N<o;ö;ó;ò,O;Ö<p,P<q,Q<r,R<s, S & SS,ß<t,T& TH, Þ &TH, þ <u;ü,U;Ü<v,V<w,W<x,X<y,Y<z,Z <"+
//                                                   "æ=œ,AE=Æ < ø,Ø < å=å,Å=Å;aa,AA < '(' < '.'< '#' < '\"' < 0<1<2<3<4<5<6<7<8<9<«<» ");
                deNoCollator.setDecomposition(Collator.NO_DECOMPOSITION);
            } catch (ParseException ex) {
                ex.printStackTrace();
            }
        }
        return (Collator) deNoCollator.clone();
    }

    public synchronized Map<String, List<String>> getNorwegianTranslationItems() {
        return norwegianTranslationItems;
    }

    public synchronized Map<String, List<String>> getGermanTranslationItems() {
        return germanTranslationItems;
    }
}