/*
 * History.java
 *
 * Created on September 5, 2007, 12:27 PM
 *
 */

package info.heinzelnisse.he;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

/**
 *
 * @author Heiko Klein
 */
public class History<T> {
    private List<T> itemList;
    static private Logger logger = Logger.getLogger(History.class.getName());
    
    /* position is next insert position */
    private int position;
    private int maxSize;
    final private long insertDelayMillis;
    Thread delayedInsertThread = null;
    private List<HistoryChangeListener> listeners = new ArrayList<HistoryChangeListener>();
    /**
     * Creates a new instance of History
     */
    public History(int maxSize, long insertDelayMillis) {
        this.maxSize = maxSize;
        this.insertDelayMillis = insertDelayMillis;
        itemList = new ArrayList<T>(maxSize+1);
        position = 0;
    }
    
    public void addHistoryChangeListener(HistoryChangeListener listener) {
        if (! listeners.contains(listener)) {
            listeners.add(listener);
        }
    }
    
    public void removeHistoryChangeListener(HistoryChangeListener listener) {
        listeners.remove(listener);
    }
    
    /**
     * wait for a certain delay
     * adds a item at the current position
     * removes all items behind the current search position
     * removes all very old items
     */
    public synchronized void addItem(T item) {
        final T myItem = item;
        Runnable sleepAndAdd = new Runnable() {
            public void run() {
                try {
                    Thread.sleep(insertDelayMillis);
                    addItemWithoutDelay(myItem);
                } catch (InterruptedException ex) {
                    // this will happen regularly
                }
            }
        };
        interruptDelayThread();
        delayedInsertThread = new Thread(sleepAndAdd);
        delayedInsertThread.start();
    }
    
    private void interruptDelayThread() {
        if (delayedInsertThread != null && delayedInsertThread.isAlive()) {
            delayedInsertThread.interrupt();
        }
    }
    
    private synchronized void addItemWithoutDelay(T item) {
        if (item.equals(getCurrent())) {
            return;
        }
        int oldPosition = position;
        if (position >= itemList.size()) {
            itemList.add(item);
            position = itemList.size();
        } else {
            itemList.set(position, item);
            position++;
        }
        if (itemList.size() > position) {
            itemList = itemList.subList(0, position - 1);
        }
        if (itemList.size() > maxSize) {
            itemList = itemList.subList(itemList.size() - maxSize, itemList.size()-1);
        }
        fireHistoryChange(oldPosition, position);
    }
    
    public synchronized T goBack() {
        interruptDelayThread();
        if (position > 1) {
            int oldPosition = position;
            position--;
            fireHistoryChange(oldPosition, position);
            return getCurrent();
        }
        return null;
    }
    
    public synchronized T goForward() {
        interruptDelayThread();
        if (position < itemList.size()) {
            int oldPosition = position;
            position++;
            fireHistoryChange(oldPosition, position);
            return getCurrent();
        }
        return null;
    }
    
    public synchronized T getCurrent() {
        int getPosition = position - 1;
        T item = null;
        try {
            item = itemList.get(getPosition);
        } catch (IndexOutOfBoundsException ioobe) {
            // don't care
        }
        return item;
    }
    
    public synchronized boolean hasNext() {
        return itemList.size() > position;
    }
    
    public synchronized boolean hasPrevious() {
        return position > 1;
    }
    
    private void fireHistoryChange(int oldPosition, int position) {
        logger.info("firing history change "+position+", "+oldPosition);
        if (position != oldPosition) {
            for (HistoryChangeListener listener : listeners) {
                listener.historyChanged();
            }
        }
    }
}
